"""
@INPROCEEDINGS{6853196,
  author={Shao, Yakun Sophia and Reagen, Brandon and Wei, Gu-Yeon and Brooks, David},
  booktitle={2014 ACM/IEEE 41st International Symposium on Computer Architecture (ISCA)},
  title={Aladdin: A pre-RTL, power-performance accelerator simulator enabling large design space exploration of customized architectures},
  year={2014},
  volume={},
  number={},
  pages={97-108},
  doi={10.1109/ISCA.2014.6853196}}
"""

from hwcomponents_library.base import LibraryEstimatorClassBase
from hwcomponents.scaling import *
from hwcomponents import actionDynamicEnergy


# Original CSV contents:
# tech_node,global_cycle_period,width|datawidth,energy,area,action
# 40nm,1e-9,32,0.21,2.78E+02,add|read
# 40nm,1e-9,32,0.0024,2.78E+02,leak
# 40nm,1e-9,32,0,2.78E+02,update|write
class AladdinAdder(LibraryEstimatorClassBase):
    """
    An adder from the Aladdin paper. Adds two values.

    Parameters
    ----------
    tech_node : str
        The technology node in meters.
    width : int, optional
        The width of the adder in bits. This is the number of bits of the input values.

    Attributes
    ----------
    tech_node : str
        The technology node in meters.
    width : int
        The width of the adder in bits. This is the number of bits of the input values.
    """
    component_name = ["Adder", "AladdinAdder", "IntAdder"]
    priority = 0.1

    def __init__(self, tech_node: float, width: int = 32):
        super().__init__(leak_power=2.40e-6, area=278.0e-12)
        self.tech_node: float = self.scale(
            "tech_node",
            tech_node,
            40e-9,
            tech_node_energy,
            tech_node_area,
            tech_node_leak,
        )
        self.width: int = self.scale("width", width, 32, linear, linear, linear)

    @actionDynamicEnergy
    def add(self) -> float:
        """
        Returns the energy for one addition operation in Joules.

        Returns
        -------
        float
            The energy for one addition operation in Joules.
        """
        return 0.21e-12

    @actionDynamicEnergy
    def read(self) -> float:
        """
        Returns the energy for one addition operation in Joules.

        Returns
        -------
        float
            The energy for one addition operation in Joules.
        """
        return 0.21e-12


# Original CSV contents:
# tech_node,global_cycle_period,width|datawidth,dynamic energy(pJ),area(um^2),action
# 40nm,1e-9,1,0.009,5.98E+00,read
# 40nm,1e-9,1,0,5.98E+00,write
# 40nm,1e-9,1,0,5.98E+00,leak|update
class AladdinRegister(LibraryEstimatorClassBase):
    """
    A register from the Aladdin paper. Stores a value.

    Parameters
    ----------
    tech_node : str
        The technology node in meters.
    width : int, optional
        The width of the register in bits.
    """
    component_name = ["Register", "AladdinRegister"]
    priority = 0.1

    def __init__(
        self,
        tech_node: float,
        width: int = 1,
        dynamic_energy: int = 1,
        area: int = 5.98e00,
    ):
        super().__init__(leak_power=0.0, area=5.98e-12)
        self.tech_node: float = self.scale(
            "tech_node",
            tech_node,
            40e-9,
            tech_node_energy,
            tech_node_area,
            tech_node_leak,
        )
        self.width: int = self.scale("width", width, 1, linear, linear, linear)
        self.dynamic_energy: int = self.scale(
            "dynamic_energy", dynamic_energy, 1, linear, linear, linear
        )

    @actionDynamicEnergy(bits_per_action="width")
    def read(self) -> float:
        """
        Returns the energy for one read operation in Joules.

        Parameters
        ----------
        bits_per_action : int
            The number of bits that are read.

        Returns
        -------
        float
            The energy for one read operation in Joules.
        """
        return 0.009e-12

    @actionDynamicEnergy(bits_per_action="width")
    def write(self) -> float:
        """
        Returns the energy for one write operation in Joules.

        Parameters
        ----------
        bits_per_action : int
            The number of bits that are written.

        Returns
        -------
        float
            The energy for one write operation in Joules.
        """
        return 0.009e-12


# Original CSV contents:
# tech_node,global_cycle_period,width|datawidth,energy(pJ),area(um^2),action
# 40nm,1e-9,32,0.02947,71,compare|read
# 40nm,1e-9,32,2.51E-05,71,leak
# 40nm,1e-9,32,0,71,update|write
class AladdinComparator(LibraryEstimatorClassBase):
    """
    A comparator from the Aladdin paper. Tells whether one value is greater than
    another.

    Parameters
    ----------
    tech_node : str
        The technology node in meters.
    width : int, optional
        The width of the comparator in bits.
    """
    component_name = ["Comparator", "AladdinComparator"]
    priority = 0.1

    def __init__(self, tech_node: float, width: int = 32):
        super().__init__(leak_power=2.51e-8, area=71.0e-12)
        self.tech_node: float = self.scale(
            "tech_node",
            tech_node,
            40e-9,
            tech_node_energy,
            tech_node_area,
            tech_node_leak,
        )
        self.width: int = self.scale("width", width, 32, linear, linear, linear)

    @actionDynamicEnergy
    def compare(self) -> float:
        """
        Returns the energy for one comparison operation in Joules.

        Returns
        -------
        float
            The energy for one comparison operation in Joules.
        """
        return 0.02947e-12

    @actionDynamicEnergy
    def read(self) -> float:
        """
        Returns the energy for one comparison operation in Joules.

        Returns
        -------
        float
            The energy for one comparison operation in Joules.
        """
        return 0.02947e-12


# Original CSV contents:
# tech_node,global_cycle_period,width|datawidth,width_a|datawidth_a,width_b|datawidth_b,energy(pJ),area(um^2),action
# 40nm,1e-9,32,32,32,12.68,6350,multiply|read
# 40nm,1e-9,32,32,32,0.08,6350,leak
# 40nm,1e-9,32,32,32,0,6350,update|write
class AladdinMultiplier(LibraryEstimatorClassBase):
    """
    A integer multiplier from the Aladdin paper. Multiplies two values.

    Parameters
    ----------
    tech_node : str
        The technology node in meters.
    width : int, optional
        The width of the multiplier in bits. Can not be set if width_a and width_b are
        set.
    width_a : int, optional
        The width of the first input value in bits.
    width_b : int, optional
        The width of the second input value in bits.
    """
    component_name = ["Multiplier", "AladdinMultiplier", "IntMultiplier"]
    priority = 0.1

    def __init__(
        self,
        tech_node: float,
        width: int = 32,
        width_a: int = 32,
        width_b: int = 32,
        energy: int = 12.68,
        area: int = 6350,
    ):
        super().__init__(leak_power=8.00e-5, area=6350.0e-12)
        self.tech_node: float = self.scale(
            "tech_node",
            tech_node,
            40e-9,
            tech_node_energy,
            tech_node_area,
            tech_node_leak,
        )
        if width_a != 32 and width != 32:
            raise ValueError(
                "width and width_a cannot both be set. Either set width of both inputs "
                "or width_a and width_b separately."
            )
        if width != 32 and width_b != 32:
            raise ValueError(
                "width and width_b cannot both be set. Either set width of both inputs "
                "or width_a and width_b separately."
            )
        self.width: int = self.scale(
            "width", width, 32, quadratic, quadratic, quadratic
        )
        self.width_a: int = self.scale("width_a", width_a, 32, linear, linear, linear)
        self.width_b: int = self.scale("width_b", width_b, 32, linear, linear, linear)

    @actionDynamicEnergy
    def multiply(self) -> float:
        """
        Returns the energy for one multiplication operation in Joules.

        Returns
        -------
        float
            The energy for one multiplication operation in Joules.
        """
        return 12.68e-12

    @actionDynamicEnergy
    def read(self) -> float:
        """
        Returns the energy for one read operation in Joules.

        Returns
        -------
        float
            The energy for one read operation in Joules.
        """
        return 12.68e-12


# Original CSV contents:
# tech_node,global_cycle_period,width|datawidth,energy(pJ),area(um^2),action
# 40nm,1e-9,32,0.25074,495.5,count|read
# 40nm,1e-9,32,0.0003213,495.5,leak
# 40nm,1e-9,32,0,495.5,update|write
class AladdinCounter(LibraryEstimatorClassBase):
    """
    A counter from the Aladdin paper. Increments a stored value.

    Parameters
    ----------
    tech_node : str
        The technology node in meters.
    width : int, optional
        The width of the counter in bits.
    """
    component_name = ["Counter", "AladdinCounter"]
    priority = 0.1

    def __init__(self, tech_node: float, width: int = 32):
        super().__init__(leak_power=3.21e-7, area=495.5e-12)
        self.tech_node: float = self.scale(
            "tech_node",
            tech_node,
            40e-9,
            tech_node_energy,
            tech_node_area,
            tech_node_leak,
        )
        self.width: int = self.scale("width", width, 32, linear, linear, linear)

    @actionDynamicEnergy
    def count(self) -> float:
        """
        Returns the energy for one increment operation in Joules.

        Returns
        -------
        float
            The energy for one increment operation in Joules.
        """
        return 0.25074e-12

    @actionDynamicEnergy
    def read(self) -> float:
        """
        Returns the energy for one increment operation in Joules.

        Returns
        -------
        float
            The energy for one increment operation in Joules.
        """
        return 0.25074e-12


class AladdinIntMAC(LibraryEstimatorClassBase):
    """
    A integer multiply-accumulate unit from the Aladdin paper. Multiplies two values
    and adds the result to a stored value.

    Parameters
    ----------
    tech_node : str
        The technology node in meters.
    adder_width : int, optional
        The width of the adder in bits.
    multiplier_width : int, optional
        The width of the multiplier in bits.
    """
    component_name = ["IntMAC", "AladdinIntMAC"]
    priority = 0.1

    def __init__(
        self, tech_node: float, adder_width: int = 16, multiplier_width: int = 8
    ):
        self.adder = AladdinAdder(tech_node, adder_width)
        self.multiplier = AladdinMultiplier(tech_node, multiplier_width)
        super().__init__(
            area=self.adder.area + self.multiplier.area,
            leak_power=self.adder.leak_power + self.multiplier.leak_power,
        )

    @actionDynamicEnergy
    def mac(self) -> float:
        """
        Returns the energy for one multiply-accumulate operation in Joules.

        Returns
        -------
        float
            The energy for one multiply-accumulate operation in Joules.
        """
        return self.adder.add() + self.multiplier.multiply()

    @actionDynamicEnergy
    def read(self) -> float:
        """
        Returns the energy for one multiply-accumulate operation in Joules.

        Returns
        -------
        float
            The energy for one multiply-accumulate operation in Joules.
        """
        return self.mac()

    @actionDynamicEnergy
    def compute(self) -> float:
        """
        Returns the energy for one multiply-accumulate operation in Joules.

        Returns
        -------
        float
            The energy for one multiply-accumulate operation in Joules.
        """
        return self.mac()



"""
@INPROCEEDINGS{9499856,
  author={Yuan, Geng and Behnam, Payman and Li, Zhengang and Shafiee, Ali and Lin, Sheng and Ma, Xiaolong and Liu, Hang and Qian, Xuehai and Bojnordi, Mahdi Nazm and Wang, Yanzhi and Ding, Caiwen},
  booktitle={2021 ACM/IEEE 48th Annual International Symposium on Computer Architecture (ISCA)},
  title={FORMS: Fine-grained Polarized ReRAM-based In-situ Computation for Mixed-signal DNN Accelerator},
  year={2021},
  volume={},
  number={},
  pages={265-278},
  doi={10.1109/ISCA52012.2021.00029}}
"""

from hwcomponents_library.base import LibraryEstimatorClassBase
from hwcomponents.scaling import *
from hwcomponents import actionDynamicEnergy
from .isaac import IsaacDAC


# Original CSV contents:
# tech_node,global_cycle_period,resolution,energy,area,n_instances,action
# 32nm,1e-9,4,0.22619,284.375,1,convert|read
# 32nm,1e-9,4,0,284.375,1,update|write|leak
# # Energy: 15.2*10^-3 W / (2.1*32*10^9 ADC BW) * 10 ^ 12 J->pJ
# # 15.2*10^-3 / (2.1*32*10^9) * 10 ^ 12
# # Area: 9100um^2 / 32
class FormsADC(LibraryEstimatorClassBase):
    """
    The analog-digital-converter (ADC) from the FORMS paper.

    Parameters
    ----------
    tech_node: float
        Technology node in meters.
    resolution: int
        Resolution of the ADC in bits.
    """

    def __init__(self, tech_node: float, resolution: int = 4):
        super().__init__(leak_power=0.0, area=284.375e-12)
        self.tech_node: float = self.scale(
            "tech_node",
            tech_node,
            32e-9,
            tech_node_energy,
            tech_node_area,
            tech_node_leak,
        )
        self.resolution: int = self.scale(
            "resolution", resolution, 4, pow_base(2), pow_base(2), pow_base(2)
        )

    @actionDynamicEnergy
    def convert(self) -> float:
        """
        Returns the energy of one ADC conversion in Joules

        Returns
        -------
        float
            Energy of one ADC conversion in Joules
        """
        return 0.22619e-12

    @actionDynamicEnergy
    def read(self) -> float:
        """
        Returns the energy of one ADC conversion operation in Joules

        Returns
        -------
        float
            Energy of one ADC conversion operation in Joules
        """
        return 0.22619e-12


class FormsDAC(IsaacDAC):
    pass



"""
@ARTICLE{9131838,
  author={Giterman, Robert and Shalom, Amir and Burg, Andreas and Fish, Alexander and Teman, Adam},
  journal={IEEE Solid-State Circuits Letters},
  title={A 1-Mbit Fully Logic-Compatible 3T Gain-Cell Embedded DRAM in 16-nm FinFET},
  year={2020},
  volume={3},
  number={},
  pages={110-113},
  keywords={Random access memory;FinFETs;Temperature measurement;Leakage currents;Power demand;Voltage measurement;Embedded DRAM;gain cell (GC);low voltage;retention time;SRAM},
  doi={10.1109/LSSC.2020.3006496}}
"""

from hwcomponents_library.base import LibraryEstimatorClassBase
from hwcomponents.scaling import *
from hwcomponents import actionDynamicEnergy
from hwcomponents_cacti import SRAM
from hwcomponents_library.library.aladdin import AladdinRegister, AladdinAdder


# Original CSV contents:
# tech_node,global_cycle_period,width|datawidth,depth,energy,area,action
# 16nm,1e-9,1024,1024,2641.92,131570,read
# 16nm,1e-9,1024,1024,2519.04,131570,write|update
# 16nm,1e-9,1024,1024,0.381,131570,leak
# # Read: 2.58 uW / MHz
# # Write: 2.46 uW / MHz
# # Leak + Refresh: (105uw leak) + (276uW refresh) = 381uW
# # @ARTICLE{9131838,
# #   author={Giterman, Robert and Shalom, Amir and Burg, Andreas and Fish, Alexander and Teman, Adam},
# #   journal={IEEE Solid-State Circuits Letters},
# #   title={A 1-Mbit Fully Logic-Compatible 3T Gain-Cell Embedded DRAM in 16-nm FinFET},
# #   year={2020},
# #   volume={3},
# #   number={},
# #   pages={110-113},
# #   keywords={Random access memory;FinFETs;Temperature measurement;Leakage currents;Power demand;Voltage measurement;Embedded DRAM;gain cell (GC);low voltage;retention time;SRAM},
# #   doi={10.1109/LSSC.2020.3006496}}
class RaaamEDRAM(LibraryEstimatorClassBase):
    """
    RAAAM EDRAM from Giterman et al. LSSC 2020. This is a MB-class embedded DRAM unit.

    Parameters
    ----------
    tech_node: float
        Technology node in meters.
    width: int
        Width of the eDRAM in bits. This is the width of a read/write port. Total size =
        width * depth.
    depth: int
        The number of entries in the eDRAM, each with `width` bits. Total size = width *
        depth.
    """

    def __init__(self, tech_node: float, width: int = 1024, depth: int = 1024):
        super().__init__(leak_power=3.81e-4, area=131570.0e-12)
        self.tech_node: float = self.scale(
            "tech_node",
            tech_node,
            16e-9,
            tech_node_energy,
            tech_node_area,
            tech_node_leak,
        )
        self.width: int = self.scale("width", width, 1024, linear, linear, linear)
        self.depth: int = self.scale(
            "depth", depth, 1024, linear, cacti_depth_energy, cacti_depth_energy
        )

    @actionDynamicEnergy(bits_per_action="width")
    def read(self) -> float:
        """
        Returns the energy consumed by a read operation in Joules.

        Parameters
        ----------
        bits_per_action: int
            The number of bits to read.

        Returns
        -------
        float
            The energy consumed by a read operation in Joules.
        """
        return 2641.92e-12

    @actionDynamicEnergy(bits_per_action="width")
    def write(self) -> float:
        """
        Returns the energy consumed by a write operation in Joules.

        Parameters
        ----------
        bits_per_action: int
            The number of bits to write.

        Returns
        -------
        float
            The energy consumed by a write operation in Joules.
        """
        return 2519.04e-12


class SmartBufferSRAM(LibraryEstimatorClassBase):
    """
    An SRAM with an address generator that sequentially reads addresses in the SRAM.

    Parameters
    ----------
        tech_node: The technology node in meters.
        width: The width of the read and write ports in bits. This is the number of bits
            that are accssed by any one read/write. Total size = width * depth.
        depth: The number of entries in the SRAM, each with `width` bits. Total size =
            width * depth.
        n_rw_ports: The number of read/write ports. Bandwidth will increase with more
            ports.
        n_banks: The number of banks. Bandwidth will increase with more banks.

    Attributes
    ----------
        sram: The SRAM buffer.
        address_reg: The register that holds the current address.
        delta_reg: The register that holds the increment value.
        adder: The adder that adds the increment value to the current address.
    """

    component_name = ["smart_buffer_sram", "smartbuffer_sram", "smartbuffersram"]
    priority = 0.3

    def __init__(
        self,
        tech_node: float,
        width: int,
        depth: int,
        n_rw_ports: int = 1,
        n_banks: int = 1,
    ):
        self.sram: SRAM = SRAM(
            tech_node=tech_node,
            width=width,
            depth=depth,
            n_rw_ports=n_rw_ports,
            n_banks=n_banks,
        )
        self.address_bits = max(math.ceil(math.log2(depth)), 1)
        self.width = width
        self.depth = depth
        self.n_rw_ports = n_rw_ports
        self.n_banks = n_banks

        self.address_reg = AladdinRegister(width=self.address_bits, tech_node=tech_node)
        self.delta_reg = AladdinRegister(width=self.address_bits, tech_node=tech_node)
        self.adder = AladdinAdder(width=self.address_bits, tech_node=tech_node)

        super().__init__(
            subcomponents=[
                self.sram,
                self.address_reg,
                self.delta_reg,
                self.adder,
            ]
        )

    @actionDynamicEnergy(bits_per_action="width")
    def read(self) -> float:
        """
        Returns the energy consumed by a read operation in Joules.

        Parameters
        ----------
        bits_per_action: int
            The number of bits to read.

        Returns
        -------
        float
            The energy consumed by a read operation in Joules.
        """
        print(f'My width is {self.width}')
        self.sram.read(bits_per_action=self.width)
        self.address_reg.read()
        self.delta_reg.read()
        self.adder.add()

    @actionDynamicEnergy(bits_per_action="width")
    def write(self) -> float:
        """
        Returns the energy consumed by a write operation in Joules.

        Parameters
        ----------
        bits_per_action: int
            The number of bits to write.

        Returns
        -------
        float
            The energy consumed by a write operation in Joules.
        """
        self.sram.write(bits_per_action=self.width)
        self.address_reg.write()
        self.delta_reg.read()
        self.adder.add()



"""
@INPROCEEDINGS{9586247,
  author={Song, Tao and Chen, Xiaoming and Zhang, Xiaoyu and Han, Yinhe},
  booktitle={2021 58th ACM/IEEE Design Automation Conference (DAC)},
  title={BRAHMS: Beyond Conventional RRAM-based Neural Network Accelerators Using Hybrid Analog Memory System},
  year={2021},
  volume={},
  number={},
  pages={1033-1038},
  doi={10.1109/DAC18074.2021.9586247}}
"""

from hwcomponents_library.base import LibraryEstimatorClassBase
from hwcomponents.scaling import *
from hwcomponents import actionDynamicEnergy


# Original CSV contents:
# tech_node,global_cycle_period,resolution,energy,area,action
# 40nm,1e-9,8,0.291,438,read|convert
# 40nm,1e-9,8,0,438,update|write|leak
# # H. Chen,X. Zhot,F. Zhang and Q. Li,"A >3GHz ERBW 1.1GS/S 8B Two-Sten SAR ADC
# # with,Q. Yu Recursive-Weight DAC," 2018 IEEE Symposium on VLSI Circuits,pp.
# # 97-98,2018 doi: 10.1109/VLSIC.2018.8502370.
# # Reported energy: 0.32mW @ 1.1GHz
# # SAR ADC, so the DAC does 1 8b convert for every convert
# # E/op: .32e-3 W * 1 / 1.1e9 seconds * 1e12pJ/J =
# # .32e-3 / 1.1e9 * 1e12 = 0.291pJ/convert
# # Area from chip picture:
# # (Picture was scaled when I screencapped it)
# # 1629px * 743px / (2805px * 1625px) * 75e-6m * 22e-6m
# # = 1629 * 743 / (2805 * 1625) * 75 * 22 = 438um^2
class BrahmsDAC(LibraryEstimatorClassBase):
    """
    Digital-analog converter (DAC) from the BRAHMS paper

    Parameters
    ----------
    tech_node: float
        Technology node in meters
    resolution: int
        Resolution of the DAC in bits
    """

    def __init__(self, tech_node: float, resolution: int = 8):
        super().__init__(leak_power=0.0, area=438.0e-12)
        self.tech_node: float = self.scale(
            "tech_node",
            tech_node,
            40e-9,
            tech_node_energy,
            tech_node_area,
            tech_node_leak,
        )
        self.resolution: int = self.scale(
            "resolution", resolution, 8, pow_base(2), pow_base(2), pow_base(2)
        )

    @actionDynamicEnergy
    def read(self) -> float:
        """
        Returns the energy of one DAC conversion in Joules

        Returns
        -------
        float
            Energy of one DAC conversion in Joules
        """
        return 0.291e-12

    @actionDynamicEnergy
    def convert(self) -> float:
        """
        Returns the energy of one DAC conversion in Joules

        Returns
        -------
        float
            Energy of one DAC conversion in Joules
        """
        return 0.291e-12



"""
@INPROCEEDINGS{9138916,
  author={Li, Weitao and Xu, Pengfei and Zhao, Yang and Li, Haitong and Xie, Yuan and Lin, Yingyan},
  booktitle={2020 ACM/IEEE 47th Annual International Symposium on Computer Architecture (ISCA)},
  title={Timely: Pushing Data Movements And Interfaces In Pim Accelerators Towards Local And In Time Domain},
  year={2020},
  volume={},
  number={},
  pages={832-845},
  doi={10.1109/ISCA45697.2020.00073}}
"""

from hwcomponents_library.base import LibraryEstimatorClassBase
from hwcomponents.scaling import *
from hwcomponents import actionDynamicEnergy
from .isaac import IsaacChip2ChipLink


# Original CSV contents:
# tech_node,global_cycle_period,energy,area,action
# 65nm,1e-9,0.0368,40,read|add
# 65nm,1e-9,0,40,write|update|leak
# # TIMELY says these don't contribute to area
# # Numbers from paper table II
class TimelyIAdder(LibraryEstimatorClassBase):
    """
    The current adder from the TIMELY paper. This unit will sum multiple currents into
    one.

    Parameters
    ----------
    tech_node: float
        Technology node in meters.
    """

    def __init__(self, tech_node: float):
        super().__init__(leak_power=0.0, area=40.0e-12)
        self.tech_node: float = self.scale(
            "tech_node",
            tech_node,
            65e-9,
            tech_node_energy,
            tech_node_area,
            tech_node_leak,
        )

    @actionDynamicEnergy
    def read(self) -> float:
        """
        Returns the energy used to sum two currents in Joules.

        Returns
        -------
        float
            The energy used to sum two currents in Joules.
        """

        return 0.0368e-12

    @actionDynamicEnergy
    def add(self) -> float:
        """
        Returns the energy used to sum two currents in Joules.

        Returns
        -------
        float
            The energy used to sum two currents in Joules.
        """

        return 0.0368e-12


# Original CSV contents:
# tech_node,global_cycle_period,n_instances,energy,area,action
# 65nm,1e-9,1,0.0023,5,drive|read|convert
# 65nm,1e-9,1,0,5,leak|update|write
# # Numbers from paper table II
class TimelyPSubBuf(LibraryEstimatorClassBase):
    """
    PSubBuf from the TIMELY paper. This unit will repeat & amplify an input voltage
    value.

    Parameters
    ----------
    tech_node: float
        Technology node in meters.
    """

    def __init__(self, tech_node: float):
        super().__init__(leak_power=0.0, area=5.0e-12)
        self.tech_node: float = self.scale(
            "tech_node",
            tech_node,
            65e-9,
            tech_node_energy,
            tech_node_area,
            tech_node_leak,
        )

    @actionDynamicEnergy
    def drive(self) -> float:
        """
        Returns the energy used to drive a voltage in Joules.

        Returns
        -------
        float
            The energy used to drive a voltage in Joules.
        """
        return 0.0023e-12

    @actionDynamicEnergy
    def read(self) -> float:
        """
        Returns the energy used to drive a voltage in Joules.

        Returns
        -------
        float
            The energy used to drive a voltage in Joules.
        """
        return 0.0023e-12


# Original CSV contents:
# tech_node,global_cycle_period,resolution,energy,area,action
# 65nm,1e-9,8,0.0375,240,convert|read
# 65nm,1e-9,8,0,240,write|leak|update
# # Numbers from paper table II
class TimelyDTC(LibraryEstimatorClassBase):
    """
    The digital-to-time converter (DTC) from the TIMELY paper. This unit will convert
    a digital value into a pulse width modulated (PWM) signal.

    Parameters
    ----------
    tech_node: float
        Technology node in meters.
    resolution: int
        Resolution of the DTC in bits.
    """

    def __init__(self, tech_node: float, resolution: int = 8):
        super().__init__(leak_power=0.0, area=240.0e-12)
        self.tech_node: float = self.scale(
            "tech_node",
            tech_node,
            65e-9,
            tech_node_energy,
            tech_node_area,
            tech_node_leak,
        )
        self.resolution: int = self.scale(
            "resolution", resolution, 8, pow_base(2), pow_base(2), pow_base(2)
        )

    @actionDynamicEnergy
    def convert(self) -> float:
        """
        Returns the energy used to convert a digital value into a PWM signal in Joules.

        Returns
        -------
        float
            The energy used to convert a digital value into a PWM signal in Joules.
        """
        return 0.0375e-12

    @actionDynamicEnergy
    def read(self) -> float:
        """
        Returns the energy used to convert a digital value into a PWM signal in Joules.

        Returns
        -------
        float
            The energy used to convert a digital value into a PWM signal in Joules.
        """
        return 0.0375e-12


# Original CSV contents:
# tech_node,global_cycle_period,resolution,energy,area,action
# 65nm,1e-9,8,0.145,310,convert|read
# 65nm,1e-9,8,0,310,leak|write|update
class TimelyTDC(LibraryEstimatorClassBase):
    """
    The time-to-digital converter (TDC) from the TIMELY paper. This unit will convert
    a pulse width modulated (PWM) signal into a digital value.

    Parameters
    ----------
    tech_node: float
        Technology node in meters.
    resolution: int
        Resolution of the TDC in bits.
    """

    def __init__(self, tech_node: float, resolution: int = 8):
        super().__init__(leak_power=0.0, area=310.0e-12)
        self.tech_node: float = self.scale(
            "tech_node",
            tech_node,
            65e-9,
            tech_node_energy,
            tech_node_area,
            tech_node_leak,
        )
        self.resolution: int = self.scale(
            "resolution", resolution, 8, pow_base(2), pow_base(2), pow_base(2)
        )

    @actionDynamicEnergy
    def convert(self) -> float:
        """
        Returns the energy used to convert a PWM signal into a digital value in Joules.

        Returns
        -------
        float
            The energy used to convert a PWM signal into a digital value in Joules.
        """
        return 0.145e-12

    @actionDynamicEnergy
    def read(self) -> float:
        """
        Returns the energy used to convert a PWM signal into a digital value in Joules.

        Returns
        -------
        float
            The energy used to convert a PWM signal into a digital value in Joules.
        """
        return 0.145e-12


# Original CSV contents:
# tech_node,global_cycle_period,rows,energy,area,action
# 65nm,1e-9,1,0.00062,5,read|drive|buffer
# 65nm,1e-9,1,0,5,leak|write|update
# # Numbers from paper table II
class TimelyXSubBuf(LibraryEstimatorClassBase):
    """
    The XSubBuf from the TIMELY paper. This unit will repeat & amplify an input current
    value.

    Parameters
    ----------
    tech_node: float
        Technology node in meters.
    rows: int
        Number of rows of the CiM array that are applying current to the XSubBuf. Power
        will increase with more rows.
    """

    def __init__(self, tech_node: float, rows: int = 1):
        super().__init__(leak_power=0.0, area=5.0e-12)
        self.tech_node: float = self.scale(
            "tech_node",
            tech_node,
            65e-9,
            tech_node_energy,
            tech_node_area,
            tech_node_leak,
        )
        self.rows: int = self.scale("rows", rows, 1, linear, noscale, noscale)

    @actionDynamicEnergy
    def read(self) -> float:
        """
        Returns the energy used to repeat & amplify an input current in Joules.

        Returns
        -------
        float
            The energy used to repeat & amplify an input current in Joules.
        """
        return 0.00062e-12

    @actionDynamicEnergy
    def drive(self) -> float:
        """
        Returns the energy used to repeat & amplify an input current in Joules.

        Returns
        -------
        float
            The energy used to repeat & amplify an input current in Joules.
        """
        return 0.00062e-12


# Original CSV contents:
# tech_node,global_cycle_period,energy,area,action
# 65nm,1e-9,0.0417,40,compare|read
# 65nm,1e-9,0,40,write|update|leak
# # Numbers from paper table II
class TimelyChargingComparator(LibraryEstimatorClassBase):
    """
    The charging comparator from the TIMELY paper. This unit will accumulate charge on a
    capacitor and trigger a singal once it exceed a reference charge.

    Parameters
    ----------
    tech_node: float
        Technology node in meters.
    """

    def __init__(self, tech_node: float):
        super().__init__(leak_power=0.0, area=40.0e-12)
        self.tech_node: float = self.scale(
            "tech_node",
            tech_node,
            65e-9,
            tech_node_energy,
            tech_node_area,
            tech_node_leak,
        )

    @actionDynamicEnergy
    def compare(self) -> float:
        """
        Returns the energy used to compare an input to the reference.

        Returns
        -------
        float
            The energy used to compare an input current to the reference.
        """
        return 0.0417e-12

    @actionDynamicEnergy
    def read(self) -> float:
        """
        Returns the energy used to compare an input to the reference.

        Returns
        -------
        float
            The energy used to compare an input current to the reference.
        """
        return 0.0417e-12


# Original CSV contents:
# tech_node,global_cycle_period,width|datawidth,depth,energy,area,action
# 65nm,1e-9,128,128,203.776,40,read
# 65nm,1e-9,128,128,496.624,40,write|update
# 65nm,1e-9,128,128,0,40,leak
class TimelyInputOutputBuffer(LibraryEstimatorClassBase):
    """
    The input/output buffers from the TIMELY paper. These digital buffers store inputs and outputs to the CiM arrays.

    Parameters
    ----------
    tech_node: float
        Technology node in meters.
    width : int, optional
        The width of the read/write port of the buffer in bits. Total size = width * depth.
    depth: int
        The number of entries in the buffer, each with `width` bits. Total size = width
        * depth.
    """

    def __init__(self, tech_node: float, width: int = 128, depth: int = 128):
        super().__init__(leak_power=0.0, area=40.0e-12)
        self.tech_node: float = self.scale(
            "tech_node",
            tech_node,
            65e-9,
            tech_node_energy,
            tech_node_area,
            tech_node_leak,
        )
        self.width: int = self.scale("width", width, 128, linear, linear, linear)
        self.depth: int = self.scale(
            "depth", depth, 128, linear, cacti_depth_energy, cacti_depth_energy
        )

    @actionDynamicEnergy(bits_per_action="width")
    def read(self) -> float:
        """
        Returns the energy used to read from the buffer in Joules.

        Parameters
        ----------
        bits_per_action: int
            The number of bits to read.

        Returns
        -------
        float
            The energy used to read from the buffer in Joules.
        """
        return 203.776e-12

    @actionDynamicEnergy(bits_per_action="width")
    def write(self) -> float:
        """
        Returns the energy used to write to the buffer in Joules.

        Parameters
        ----------
        bits_per_action: int
            The number of bits to write.

        Returns
        -------
        float
            The energy used to write to the buffer in Joules.
        """
        return 496.624e-12


class TimelyChip2ChipLink(IsaacChip2ChipLink):
    pass



from hwcomponents_library.base import LibraryEstimatorClassBase
from hwcomponents.scaling import *
from hwcomponents import actionDynamicEnergy


# Original CSV contents:
# global_cycle_period,energy,area,n_instances,action
# 1e-9,0,0,1,read|update|leak|write|*
class DummyStorage(LibraryEstimatorClassBase):
    """
    A dummy storage component. Has zero area, zero energy, and zero leakage power.

    Parameters
    ----------
    tech_node: float, optional
        Technology node in meters. This is not used.
    """

    def __init__(self, tech_node: float | None = None):
        super().__init__(leak_power=0.0, area=0.0)
        self.tech_node: float = tech_node

    @actionDynamicEnergy
    def read(self) -> float:
        """
        Returns the energy of one read operation in Joules. Energy is zero.

        Returns
        -------
        float
            Energy of one read operation in Joules
        """
        return 0.0

    @actionDynamicEnergy
    def write(self) -> float:
        """
        Returns the energy of one write operation in Joules. Energy is zero.

        Returns
        -------
        float
            Energy of one write operation in Joules
        """
        return 0.0


# Original CSV contents:
# global_cycle_period,energy,area,n_instances,action
# 1e-9,0,0,1,read|update|leak|write|*
class DummyCompute(LibraryEstimatorClassBase):
    """
    A dummy compute component. Has zero area, zero energy, and zero leakage power.

    Parameters
    ----------
    tech_node: float, optional
        Technology node in meters. This is not used.
    """

    def __init__(self, tech_node: float | None = None):
        super().__init__(leak_power=0.0, area=0.0)
        self.tech_node: float = tech_node

    @actionDynamicEnergy
    def read(self) -> float:
        """
        Returns the energy of one compute operation in Joules. Energy is zero.

        Returns
        -------
        float
            Energy of one compute operation in Joules
        """
        return 0.0

    @actionDynamicEnergy
    def compute(self) -> float:
        """
        Returns the energy of one compute operation in Joules. Energy is zero.

        Returns
        -------
        float
            Energy of one compute operation in Joules
        """
        return 0.0


class DummyMemory(LibraryEstimatorClassBase):
    """
    A dummy memory component. Has zero area, zero energy, and zero leakage power.

    Parameters
    ----------
    tech_node: float
        Technology node in meters. This is not used.
    """

    def __init__(self, tech_node: float | None = None):
        super().__init__(leak_power=0.0, area=0.0)
        self.tech_node: float = tech_node

    @actionDynamicEnergy
    def read(self) -> float:
        """
        Returns the energy of one read operation in Joules. Energy is zero.

        Returns
        -------
        float
            Energy of one read operation in Joules
        """
        return 0.0

    @actionDynamicEnergy
    def write(self) -> float:
        """
        Returns the energy of one write operation in Joules. Energy is zero.

        Returns
        -------
        float
            Energy of one write operation in Joules
        """
        return 0.0


class DummyNetwork(LibraryEstimatorClassBase):
    """
    A dummy network component. Has zero area, zero energy, and zero leakage power.

    Parameters
    ----------
    tech_node: float
        Technology node in meters. This is not used.
    """

    def __init__(self, tech_node: float | None = None):
        super().__init__(leak_power=0.0, area=0.0)
        self.tech_node: float = tech_node

    @actionDynamicEnergy
    def read(self) -> float:
        """
        Returns the energy of one read operation in Joules. Energy is zero.

        Returns
        -------
        float
            Energy of one read operation in Joules
        """
        return 0.0

    @actionDynamicEnergy
    def write(self) -> float:
        """
        Returns the energy of one write operation in Joules. Energy is zero.

        Returns
        -------
        float
            Energy of one write operation in Joules
        """
        return 0.0



"""Setup.py to ensure hwcomponents._version_scheme is importable during build."""
import sys
from pathlib import Path

# Add paths to ensure hwcomponents can be imported
current_dir = Path(__file__).parent.absolute()
if str(current_dir) not in sys.path:
    sys.path.insert(0, str(current_dir))

# Try to find hwcomponents in parent directories (for submodule case)
parent_dir = current_dir.parent.parent
hwcomponents_dir = parent_dir / "hwcomponents"
if hwcomponents_dir.exists() and str(parent_dir) not in sys.path:
    sys.path.insert(0, str(parent_dir))

from setuptools import setup

setup()

